from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

import mujoco
import numpy as np

from ..exceptions import InvalidMocapBody
from .base import MatrixLieGroup
from .so3 import SO3
from .utils import get_epsilon, skew

_IDENTITY_WXYZ_XYZ = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float64)


@dataclass(frozen=True)
class SE3(MatrixLieGroup):
    """Special Euclidean group for proper rigid transforms in 3D.

    Internal parameterization is (qw, qx, qy, qz, x, y, z). Tangent parameterization is
    (vx, vy, vz, omega_x, omega_y, omega_z).
    """

    wxyz_xyz: np.ndarray
    matrix_dim: int = 4
    parameters_dim: int = 7
    tangent_dim: int = 6
    space_dim: int = 3

    def __repr__(self) -> str:
        quat = np.round(self.wxyz_xyz[:4], 5)
        xyz = np.round(self.wxyz_xyz[4:], 5)
        return f"{self.__class__.__name__}(wxyz={quat}, xyz={xyz})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SE3):
            return NotImplemented
        return np.array_equal(self.wxyz_xyz, other.wxyz_xyz)

    def copy(self) -> SE3:
        return SE3(wxyz_xyz=np.array(self.wxyz_xyz))

    def parameters(self) -> np.ndarray:
        return self.wxyz_xyz

    @classmethod
    def identity(cls) -> SE3:
        return SE3(wxyz_xyz=_IDENTITY_WXYZ_XYZ)

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        translation: np.ndarray,
    ) -> SE3:
        assert translation.shape == (SE3.space_dim,)
        return SE3(wxyz_xyz=np.concatenate([rotation.wxyz, translation]))

    @classmethod
    def from_rotation(cls, rotation: SO3) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=rotation, translation=np.zeros(SE3.space_dim)
        )

    @classmethod
    def from_translation(cls, translation: np.ndarray) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.identity(), translation=translation
        )

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> SE3:
        assert matrix.shape == (SE3.matrix_dim, SE3.matrix_dim)
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            translation=matrix[:3, 3],
        )

    @classmethod
    def from_mocap_id(cls, data: mujoco.MjData, mocap_id: int) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3(data.mocap_quat[mocap_id]),
            translation=data.mocap_pos[mocap_id],
        )

    @classmethod
    def from_mocap_name(
        cls, model: mujoco.MjModel, data: mujoco.MjData, mocap_name: str
    ) -> SE3:
        mocap_id = model.body(mocap_name).mocapid[0]
        if mocap_id == -1:
            raise InvalidMocapBody(mocap_name, model)
        return SE3.from_mocap_id(data, mocap_id)

    @classmethod
    def sample_uniform(cls) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.sample_uniform(),
            translation=np.random.uniform(-1.0, 1.0, size=(SE3.space_dim,)),
        )

    def rotation(self) -> SO3:
        return SO3(wxyz=self.wxyz_xyz[:4])

    def translation(self) -> np.ndarray:
        return self.wxyz_xyz[4:]

    def as_matrix(self) -> np.ndarray:
        hmat = np.eye(self.matrix_dim, dtype=np.float64)
        hmat[:3, :3] = self.rotation().as_matrix()
        hmat[:3, 3] = self.translation()
        return hmat

    @classmethod
    def exp(cls, tangent: np.ndarray) -> SE3:
        assert tangent.shape == (cls.tangent_dim,)
        rotation = SO3.exp(tangent[3:])
        theta = np.float64(mujoco.mju_norm3(tangent[3:]))
        t2 = theta * theta
        if t2 < get_epsilon(t2.dtype):
            v_mat = rotation.as_matrix()
        else:
            skew_omega = skew(tangent[3:])
            v_mat = (
                np.eye(3, dtype=np.float64)
                + (1.0 - np.cos(theta)) / t2 * skew_omega
                + (theta - np.sin(theta)) / (t2 * theta) * (skew_omega @ skew_omega)
            )
        return cls.from_rotation_and_translation(
            rotation=rotation,
            translation=v_mat @ tangent[:3],
        )

    def inverse(self) -> SE3:
        inverse_wxyz_xyz = np.empty(SE3.parameters_dim, dtype=np.float64)
        mujoco.mju_negQuat(inverse_wxyz_xyz[:4], self.wxyz_xyz[:4])
        mujoco.mju_rotVecQuat(
            inverse_wxyz_xyz[4:], -1.0 * self.wxyz_xyz[4:], inverse_wxyz_xyz[:4]
        )
        return SE3(wxyz_xyz=inverse_wxyz_xyz)

    def normalize(self) -> SE3:
        normalized_wxyz_xyz = np.array(self.wxyz_xyz)
        mujoco.mju_normalize4(normalized_wxyz_xyz[:4])
        return SE3(wxyz_xyz=normalized_wxyz_xyz)

    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SE3.space_dim,)
        rotated_target = np.empty(SE3.space_dim, dtype=np.float64)
        mujoco.mju_rotVecQuat(rotated_target, target, self.wxyz_xyz[:4])
        return rotated_target + self.wxyz_xyz[4:]

    def multiply(self, other: SE3) -> SE3:
        wxyz_xyz = np.empty(SE3.parameters_dim, dtype=np.float64)
        mujoco.mju_mulQuat(wxyz_xyz[:4], self.wxyz_xyz[:4], other.wxyz_xyz[:4])
        mujoco.mju_rotVecQuat(wxyz_xyz[4:], other.wxyz_xyz[4:], self.wxyz_xyz[:4])
        wxyz_xyz[4:] += self.wxyz_xyz[4:]
        return SE3(wxyz_xyz=wxyz_xyz)

    def log(self) -> np.ndarray:
        omega = self.rotation().log()
        theta = np.float64(mujoco.mju_norm3(omega))
        t2 = theta * theta
        skew_omega = skew(omega)
        skew_omega2 = skew_omega @ skew_omega
        if t2 < get_epsilon(t2.dtype):
            vinv_mat = (
                np.eye(3, dtype=np.float64) - 0.5 * skew_omega + skew_omega2 / 12.0
            )
        else:
            half_theta = 0.5 * theta
            vinv_mat = (
                np.eye(3, dtype=np.float64)
                - 0.5 * skew_omega
                + (1.0 - 0.5 * theta * np.cos(half_theta) / np.sin(half_theta))
                / t2
                * skew_omega2
            )
        tangent = np.empty(SE3.tangent_dim, dtype=np.float64)
        tangent[:3] = vinv_mat @ self.translation()
        tangent[3:] = omega
        return tangent

    def adjoint(self) -> np.ndarray:
        rotation = self.rotation()
        rotation_mat = rotation.as_matrix()
        tangent_mat = skew(self.translation()) @ rotation_mat
        adjoint_mat = np.zeros((SE3.tangent_dim, SE3.tangent_dim), dtype=np.float64)
        adjoint_mat[:3, :3] = rotation_mat
        adjoint_mat[:3, 3:] = tangent_mat
        adjoint_mat[3:, 3:] = rotation_mat
        return adjoint_mat

    # Jacobians.

    # Eqn 179 a)
    @classmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        theta_squared = np.float64(mujoco.mju_dot3(other[3:], other[3:]))
        if theta_squared < get_epsilon(theta_squared.dtype):
            return np.eye(cls.tangent_dim)
        ljac_se3 = np.zeros((cls.tangent_dim, cls.tangent_dim), dtype=np.float64)
        ljac_translation = _getQ(other)
        ljac_so3 = SO3.ljac(other[3:])
        ljac_se3[:3, :3] = ljac_so3
        ljac_se3[:3, 3:] = ljac_translation
        ljac_se3[3:, 3:] = ljac_so3
        return ljac_se3

    # Eqn 179 b)
    @classmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        theta_squared = np.float64(mujoco.mju_dot3(other[3:], other[3:]))
        if theta_squared < get_epsilon(theta_squared.dtype):
            return np.eye(cls.tangent_dim)
        ljacinv_se3 = np.zeros((cls.tangent_dim, cls.tangent_dim), dtype=np.float64)
        ljac_translation = _getQ(other)
        ljacinv_so3 = SO3.ljacinv(other[3:])
        ljacinv_se3[:3, :3] = ljacinv_so3
        ljacinv_se3[:3, 3:] = -ljacinv_so3 @ ljac_translation @ ljacinv_so3
        ljacinv_se3[3:, 3:] = ljacinv_so3
        return ljacinv_se3

    def clamp(
        self,
        x_translation: Tuple[float, float] = (-np.inf, np.inf),
        y_translation: Tuple[float, float] = (-np.inf, np.inf),
        z_translation: Tuple[float, float] = (-np.inf, np.inf),
        roll_radians: Tuple[float, float] = (-np.inf, np.inf),
        pitch_radians: Tuple[float, float] = (-np.inf, np.inf),
        yaw_radians: Tuple[float, float] = (-np.inf, np.inf),
    ) -> SE3:
        """Clamp a SE3 within translation and RPY limits.

        Args:
            x_translation: The (lower, upper) limits for translation along the x axis.
            y_translation: The (lower, upper) limits for translation along the y axis.
            z_translation: The (lower, upper) limits for translation along the z axis.
            roll_radians: The (lower, upper) limits for the roll.
            pitch_radians: The (lower, upper) limits for the pitch.
            yaw_radians: The (lower, upper) limits for the yaw.

        Returns:
            A SE3 within the translation and RPY limits.
        """
        translation_limits = np.array([x_translation, y_translation, z_translation])
        return SE3.from_rotation_and_translation(
            rotation=self.rotation().clamp(roll_radians, pitch_radians, yaw_radians),
            translation=np.clip(
                self.translation(), translation_limits[:, 0], translation_limits[:, 1]
            ),
        )


# Eqn 180.
def _getQ(c) -> np.ndarray:
    theta = np.float64(mujoco.mju_norm3(c[3:]))
    t2 = theta * theta
    A = 0.5
    if t2 < get_epsilon(t2.dtype):
        B = (1.0 / 6.0) + (1.0 / 120.0) * t2
        C = -(1.0 / 24.0) + (1.0 / 720.0) * t2
        D = -(1.0 / 60.0)
    else:
        t4 = t2 * t2
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        B = (theta - sin_theta) / (t2 * theta)
        C = (1.0 - 0.5 * t2 - cos_theta) / t4
        D = (2.0 * theta - 3.0 * sin_theta + theta * cos_theta) / (2.0 * t4 * theta)
    V = skew(c[:3])
    W = skew(c[3:])
    VW = V @ W
    WV = VW.T
    WVW = WV @ W
    VWW = VW @ W
    return (
        +A * V
        + B * (WV + VW + WVW)
        - C * (VWW - VWW.T - 3.0 * WVW)
        + D * (WVW @ W + W @ WVW)
    )



"""Kinematic tasks."""

from .com_task import ComTask
from .damping_task import DampingTask
from .equality_constraint_task import EqualityConstraintTask
from .frame_task import FrameTask
from .kinetic_energy_regularization_task import KineticEnergyRegularizationTask
from .posture_task import PostureTask
from .relative_frame_task import RelativeFrameTask
from .task import BaseTask, Objective, Task

__all__ = (
    "BaseTask",
    "ComTask",
    "FrameTask",
    "Objective",
    "DampingTask",
    "PostureTask",
    "RelativeFrameTask",
    "Task",
    "EqualityConstraintTask",
    "KineticEnergyRegularizationTask",
)



"""Kinetic energy regularization task implementation."""

from __future__ import annotations

from typing import Optional, SupportsFloat

import numpy as np

from ..configuration import Configuration
from ..exceptions import IntegrationTimestepNotSet, TaskDefinitionError
from .task import BaseTask, Objective


class KineticEnergyRegularizationTask(BaseTask):
    r"""Kinetic-energy regularization.

    This task, often used with a low priority in the task stack, penalizes the system's
    kinetic energy. Formally, it contributes the following term to the quadratic
    program:

    .. math::
        \frac{1}{2}\, \lambda\, \Delta \mathbf{q}^\top M(\mathbf{q})\,\Delta \mathbf{q},

    where :math:`\Delta \mathbf{q}\in\mathbb{R}^{n_v}` is the vector of joint
    displacements, :math:`M(\mathbf{q})` is the joint-space inertia matrix, and
    :math:`\lambda` is the scalar strength of the regularization.

    .. note::

        This task can be seen as an inertia-weighted version of the
        :class:`~.DampingTask`. Degrees of freedom with higher inertia will move less
        for the same cost.

    .. warning::

        The integration timestep :math:`\Delta t` must be set via :meth:`set_dt`
        before use. This ensures the cost is expressed in units of energy (Joules).

    Example:

    .. code-block:: python

        task = KineticEnergyRegularizationTask(cost=1e-4)
        task.set_dt(0.02)
    """

    def __init__(self, cost: SupportsFloat):
        cost = float(cost)
        if cost < 0:
            raise TaskDefinitionError(f"{self.__class__.__name__} cost should be >= 0")
        self.cost: float = cost

        # Kinetic energy is defined as T = ½ * q̇ᵀ * M * q̇. Since q̇ ≈ Δq / dt, we
        # substitute: T ≈ ½ * (Δq / dt)ᵀ * M * (Δq / dt) = ½ * Δqᵀ * (M / dt²) * Δq.
        # Therefore, we scale the inertia matrix by 1 / dt² to penalize energy in
        # terms of joint displacements Δq.
        self.inv_dt_sq: Optional[float] = None

    def set_dt(self, dt: float) -> None:
        """Set the integration timestep.

        Args:
            dt: Integration timestep in [s].
        """
        self.inv_dt_sq = 1.0 / dt**2

    def compute_qp_objective(self, configuration: Configuration) -> Objective:
        if self.inv_dt_sq is None:
            raise IntegrationTimestepNotSet(self.__class__.__name__)
        M = configuration.get_inertia_matrix()
        H = self.cost * self.inv_dt_sq * M
        c = np.zeros(configuration.nv)
        return Objective(H, c)



"""Center-of-mass task implementation."""

from __future__ import annotations

from typing import Optional

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..exceptions import InvalidTarget, TargetNotSet, TaskDefinitionError
from .task import Task


class ComTask(Task):
    """Regulate the center-of-mass (CoM) of a robot.

    Attributes:
        target_com: Target position of the CoM.

    Example:

    .. code-block:: python

        com_task = ComTask(model, cost=1.0)

        # Update the target CoM directly.
        com_desired = np.zeros(3)
        com_task.set_target(com_desired)

        # Or from a keyframe defined in the model.
        configuration.update_from_keyframe("home")
        com_task.set_target_from_configuration(configuration)
    """

    k: int = 3
    target_com: Optional[np.ndarray]

    def __init__(
        self,
        cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(cost=np.zeros((self.k,)), gain=gain, lm_damping=lm_damping)
        self.target_com = None

        self.set_cost(cost)

    def set_cost(self, cost: npt.ArrayLike) -> None:
        """Set the cost of the CoM task.

        Args:
            cost: A vector of shape (1,) (aka identical cost for all coordinates),
                or (3,) (aka different costs for each coordinate).
        """
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self.k):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"(aka identical cost for all coordinates) or ({self.k},). "
                f"Got {cost.shape}"
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost must be >= 0")
        self.cost[:] = cost

    def set_target(self, target_com: npt.ArrayLike) -> None:
        """Set the target CoM position in the world frame.

        Args:
            target_com: A vector of shape (3,) representing the desired
                center-of-mass position in the world frame.
        """
        target_com = np.atleast_1d(target_com)
        if target_com.ndim != 1 or target_com.shape[0] != (self.k):
            raise InvalidTarget(
                f"Expected target CoM to have shape ({self.k},) but got "
                f"{target_com.shape}"
            )
        self.target_com = target_com.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target CoM from a given robot configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(configuration.data.subtree_com[1])

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the CoM task error.

        The center of mass :math:`c(q)` for a collection of bodies :math:`\mathcal{B}`
        is the mass-weighted average of their individual centers of mass. After running
        forward kinematics, in particular after calling ``mj_comPos``, MuJoCo stores
        the CoM of each subtree in ``data.subtree_com``. This task uses the CoM of the
        subtree starting from body 1, which is the entire robot excluding the world
        body (body 0).

        The task error :math:`e(q)` is the difference between the current CoM
        :math:`c(q)` and the target CoM :math:`c^*`:

        .. math::

            e(q) = c(q) - c^*

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Center-of-mass task error vector :math:`e(q)`.
        """
        if self.target_com is None:
            raise TargetNotSet(self.__class__.__name__)
        # Note: body 0 is the world body, so we start from body 1 (the robot).
        # TODO(kevin): Don't hardcode subtree index.
        return configuration.data.subtree_com[1] - self.target_com

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the Jacobian of the CoM task error :math:`e(q)`.

        The Jacobian is the derivative of this error with respect to the
        generalized coordinates :math:`q`. Since the target :math:`c^*` is
        constant, the Jacobian of the error simplifies to the Jacobian of the
        CoM position :math:`c(q)`:

        .. math::

            J(q) = \frac{\partial e(q)}{\partial q} = \frac{\partial c(q)}{\partial q}

        MuJoCo's ``mj_jacSubtreeCom`` function computes this Jacobian using the
        formula:

        .. math::

            \frac{\partial c(q)}{\partial q} =
            \frac{1}{M} \sum_{i \in \mathcal{B}} m_i \frac{\partial p_i(q)}{\partial q}
            = \frac{1}{M} \sum_{i \in \mathcal{B}} m_i J_i(q)

        where :math:`M = \sum_{i \in \mathcal{B}} m_i` is the total mass of the subtree,
        :math:`m_i` is the mass of body :math:`i`, :math:`p_i(q)` is the position
        of the origin of body frame :math:`i` in the world frame, :math:`J_i(q) =
        \frac{\partial p_i(q)}{\partial q}` is the Jacobian mapping joint velocities to
        the linear velocity of the origin of body frame :math:`i`, and the sum is over
        all bodies :math:`\mathcal{B}` in the specified subtree (body 1 and its
        descendants).

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Jacobian of the center-of-mass task error :math:`J(q)`.
        """
        # NOTE: We don't need a target CoM to compute this Jacobian.
        jac = np.empty((self.k, configuration.nv))
        mujoco.mj_jacSubtreeCom(configuration.model, configuration.data, jac, 1)
        return jac



"""Damping task implementation."""

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from .posture_task import PostureTask


class DampingTask(PostureTask):
    r"""L2-regularization on joint velocities (a.k.a. *velocity damping*).

    This task, typically used with a low priority in the task stack, adds a
    Levenberg-Marquardt term to the quadratic program, favoring **minimum-norm
    joint velocities** in redundant or near-singular situations. Formally, it
    contributes the following term to the quadratic program:

    .. math::
        \frac{1}{2}\,\Delta \mathbf{q}^\top \Lambda\,\Delta \mathbf{q},

    where :math:`\Delta \mathbf{q}\in\mathbb{R}^{n_v}` is the vector of joint
    displacements and :math:`\Lambda = \mathrm{diag}(\lambda_1, \ldots, \lambda_{n_v})`
    is a diagonal matrix of per-DoF damping weights specified via ``cost``. A larger
    :math:`\lambda_i` reduces motion in DoF :math:`i`. With no other active
    tasks, the robot remains at rest. Unlike the `damping` parameter in
    :func:`~.solve_ik`, which is uniformly applied to all DoFs, this task does
    not affect the floating-base coordinates.

    .. note::

        This task does not favor a particular posture, only small instantaneous motion.
        If you need a posture bias, use :class:`~.PostureTask` instead.

    Example:

    .. code-block:: python

        # Uniform damping across all degrees of freedom.
        damping_task = DampingTask(model, cost=1.0)

        # Custom damping.
        cost = np.zeros(model.nv)
        cost[:3] = 1.0  # High damping for the first 3 joints.
        cost[3:] = 0.1  # Lower damping for the remaining joints.
        damping_task = DampingTask(model, cost)
    """

    def __init__(self, model: mujoco.MjModel, cost: npt.ArrayLike):
        super().__init__(model, cost, gain=0.0, lm_damping=0.0)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the damping task error.

        The damping task does not chase a reference; its desired joint velocity
        is identically **zero**, so the task error is always:

        .. math:: e = \mathbf 0 \in \mathbb R^{n_v}.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Zero vector of length :math:`n_v`.
        """
        return np.zeros(configuration.nv)



"""Posture task implementation."""

from __future__ import annotations

from typing import Optional

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..exceptions import InvalidTarget, TargetNotSet, TaskDefinitionError
from ..utils import get_freejoint_dims
from .task import Task


class PostureTask(Task):
    """Regulate joint angles towards a target posture.

    Often used with a low priority in the task stack, this task acts like a regularizer,
    biasing the solution toward a specific joint configuration.

    Attributes:
        target_q: Target configuration :math:`q^*`, of shape :math:`(n_q,)`. Units are
            radians for revolute joints and meters for prismatic joints. Note that
            floating-base coordinates are not affected by this task but should be
            included in the target configuration.

    Example:

    .. code-block:: python

        posture_task = PostureTask(model, cost=1e-3)

        # Update the target posture directly.
        q_desired = ...
        posture_task.set_target(q_desired)

        # Or from a keyframe defined in the model.
        configuration.update_from_keyframe("home")
        posture_task.set_target_from_configuration(configuration)
    """

    target_q: Optional[np.ndarray]

    def __init__(
        self,
        model: mujoco.MjModel,
        cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(
            cost=np.zeros((model.nv,)),
            gain=gain,
            lm_damping=lm_damping,
        )
        self.target_q = None

        self._v_ids: np.ndarray | None
        _, v_ids_or_none = get_freejoint_dims(model)
        if v_ids_or_none:
            self._v_ids = np.asarray(v_ids_or_none)
        else:
            self._v_ids = None

        self.k = model.nv
        self.nq = model.nq
        self.set_cost(cost)

    def set_cost(self, cost: npt.ArrayLike) -> None:
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self.k):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"(aka identical cost for all dofs) or ({self.k},). Got {cost.shape}"
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost should be >= 0")
        self.cost[: self.k] = cost

    def set_target(self, target_q: npt.ArrayLike) -> None:
        """Set the target posture.

        Args:
            target_q: A vector of shape (nq,) representing the desired joint
                configuration.
        """
        target_q = np.atleast_1d(target_q)
        if target_q.ndim != 1 or target_q.shape[0] != (self.nq):
            raise InvalidTarget(
                f"Expected target posture to have shape ({self.nq},) but got "
                f"{target_q.shape}"
            )
        self.target_q = target_q.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target posture by extracting it from the current configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(configuration.q)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the posture task error.

        The error is defined as:

        .. math::

            e(q) = q^* \ominus q

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task error vector :math:`e(q)`.
        """
        if self.target_q is None:
            raise TargetNotSet(self.__class__.__name__)

        # NOTE: mj_differentiatePos calculates qpos2 ⊖ qpos1.
        qvel = np.empty(configuration.nv)
        mujoco.mj_differentiatePos(
            m=configuration.model,
            qvel=qvel,
            dt=1.0,
            qpos1=self.target_q,
            qpos2=configuration.q,
        )

        if self._v_ids is not None:
            qvel[self._v_ids] = 0.0

        return qvel

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the posture task Jacobian.

        The task Jacobian is defined as:

        .. math::

            J(q) = I_{n_v}

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Posture task jacobian :math:`J(q)`.
        """
        jac = np.eye(configuration.nv)
        if self._v_ids is not None:
            jac[:, self._v_ids] = 0.0
        return jac
